#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import asyncio
from typing import Any

from django.conf import settings
from django.utils.module_loading import import_string
from taskiq import AsyncBroker

from unikit.contrib.django.di.base import BaseDiSupportedApp
from unikit.contrib.taskiq.service import TaskiqWorkerService
from unikit.di import root_container
from unikit.utils.list_utils import ensure_list
from unikit.utils.logger import LogMixin
from unikit.worker import WorkerServiceRegistry


class TaskiqConfig(BaseDiSupportedApp, LogMixin):
    """
    TaskiqConfig is a Django application configuration for Taskiq.

    Supported Settings:
       - TASKIQ_BROKER_DEFINITION: str - Dotted path to the broker object instance to use,
        if ommited the new one will be created.
       - TASKIQ_BROKER_URL: str - URL to the broker.
       - TASKIQ_TASK_DISCOVERY_PATTERN: str - Glob pattern to discover tasks, default: `**/tasks.py`.
       - TASKIQ_RESULT_BACKEND_URL: str - URL to the result backend.
       - TASKIQ_RESULT_BACKEND_URL: str - URL to the result backend.
    """

    name = "unikit.contrib.django.taskiq"
    label = "django_taskiq"
    default = True

    def __init__(self, app_name: str, app_module: Any) -> None:
        super().__init__(app_name, app_module)
        self.broker_paths: dict[str, str] = {}
        self.brokers: dict[str, AsyncBroker] = {}
        self.task_discovery_pattern = "**/tasks.py"

    @property
    def default_broker_name(self) -> str:
        """Default broker name."""
        return next(iter(self.broker_paths))

    @property
    def default_broker_path(self) -> str:
        """Default broker path."""
        return self.broker_paths[self.default_broker_name]

    @property
    def default_broker(self) -> AsyncBroker:
        """Default broker instance."""
        return self.brokers[self.default_broker_name]

    def ready(self) -> None:
        """Django callback invoked when application is ready to be used."""
        super().ready()

        self.task_discovery_pattern = getattr(settings, "TASKIQ_TASK_DISCOVERY_PATTERN", self.task_discovery_pattern)
        init_brokers_automatically = getattr(settings, "TASKIQ_AUTO_INIT_BROKERS", True)

        broker_object_paths = getattr(settings, "TASKIQ_BROKER_DEFINITION", None)
        if broker_object_paths is not None:
            broker_object_paths = ensure_list(broker_object_paths)
            for bp in broker_object_paths:
                broker: AsyncBroker = import_string(bp)
                _, broker_name = bp.rsplit(".", 1)
                self.broker_paths[broker_name] = bp
                self.brokers[broker_name] = broker

        if not self.brokers:
            raise ValueError("TASKIQ_BROKER_OBJECT is not set, please provide a broker object.")

        broker_registry = root_container.get(WorkerServiceRegistry)
        for name, b in self.brokers.items():
            broker_registry.register(name, TaskiqWorkerService(b))

        if init_brokers_automatically:
            self.init_brokers()

    async def ainit_brokers(self) -> None:
        """Initialize all brokers."""
        await asyncio.gather(*[b.startup() for b in self.brokers.values()])

    def init_brokers(self) -> None:
        """Initialize all brokers."""
        self._get_event_loop().run_until_complete(self.ainit_brokers())

    def _get_event_loop(self) -> asyncio.AbstractEventLoop:
        """Get event loop."""
        from django.apps import apps

        for app in apps.get_app_configs():
            if app.name == "daphne":
                try:
                    from daphne.server import twisted_loop  # type: ignore

                    return twisted_loop
                except ImportError:
                    pass
        return asyncio.get_event_loop()



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import datetime
import logging
import string
from typing import Self

from unikit.lock.dto import Lock
from unikit.utils.type_utils import none_raises

logger = logging.getLogger(__name__)

DEFAULT_LOCK_TIMEOUT = datetime.timedelta(minutes=5)
CHARACTERS = string.ascii_letters + string.digits


class RedisLock(Lock):
    """Redis lock object."""

    def __init__(
        self,
        lock_id: str,
        op_name: str | None,
        target_id: str | None,
        ts_acquired: datetime.datetime | None = None,
        ts_expires: datetime.datetime | None = None,
        lock_owner: str | None = None,
        secret: str | None = None,
    ) -> None:
        super().__init__()
        self.__secret = secret
        self.__lock_id = lock_id
        self.__op_name = op_name
        self.__lock_owner = lock_owner
        self.__target_id = target_id
        self.__ts_acquired = ts_acquired or datetime.datetime.now(tz=datetime.timezone.utc)
        self.__ts_expires = ts_expires or self.__ts_acquired + DEFAULT_LOCK_TIMEOUT

    def __str__(self) -> str:
        data = f"ObjectLock({self.get_target_id()}" if self.get_target_id() else f"OpLock({self.get_operation_name()}"
        data += f", expires at {self.get_ts_expires().isoformat()})"
        return data

    @classmethod
    def from_dict(cls, d: dict[str, str | None]) -> Self:
        """Create a lock object from a dictionary."""
        try:
            return cls(
                lock_id=none_raises(d["id"]),
                op_name=d.get("op"),
                target_id=d.get("t"),
                ts_acquired=datetime.datetime.fromisoformat(none_raises(d["acq"])),
                ts_expires=datetime.datetime.fromisoformat(none_raises(d["exp"])),
                secret=d.get("s"),
                lock_owner=d.get("own"),
            )
        except KeyError as e:
            raise ValueError(f"Invalid redis lock object. Missing field {e}. Object: {d}")
        except ValueError as e:
            raise ValueError(f"Invalid redis lock object. Invalid timestamp: {e}") from e

    def get_lock_id(self) -> str:
        """Get the lock ID."""
        return self.__lock_id

    def get_operation_name(self) -> str | None:
        """Get the operation name."""
        return self.__op_name

    def get_target_id(self) -> str | None:
        """Get the target ID."""
        return self.__target_id

    def get_ts_acquired(self) -> datetime.datetime:
        """Get the timestamp when the lock was acquired."""
        return self.__ts_acquired

    def get_ts_expires(self) -> datetime.datetime:
        """Get the timestamp when the lock expires."""
        return self.__ts_expires

    def get_secret(self) -> str:
        """Get the lock secret."""
        return none_raises(self.__secret)

    def get_owner(self) -> str | None:
        """Get the lock owner."""
        return self.__lock_owner

    def to_dict(self) -> dict[str, str | None]:
        """Convert the lock object to a dictionary."""
        return dict(
            id=self.get_lock_id(),
            op=self.get_operation_name(),
            t=self.get_target_id(),
            acq=self.get_ts_acquired().isoformat(),
            exp=self.get_ts_expires().isoformat(),
            s=self.__secret,
            own=self.__lock_owner,
        )

    def clone(
        self,
        ts_acquired: datetime.datetime | None = None,
        ts_expires: datetime.datetime | None = None,
    ) -> Self:
        """Clone the lock object with new timestamps."""
        return self.__class__(
            lock_id=self.get_lock_id(),
            op_name=self.get_operation_name(),
            target_id=self.get_target_id(),
            ts_acquired=ts_acquired or self.get_ts_acquired(),
            ts_expires=ts_expires or self.get_ts_expires(),
            lock_owner=self.get_owner(),
            secret=self.get_secret(),
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RedisLock):
            return False
        return self.get_lock_id() == other.get_lock_id() and self.get_ts_acquired() == other.get_ts_acquired()



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import datetime
import json
from typing import Any

from redis.asyncio import StrictRedis

from unikit.contrib.redis.lock.base import BaseRedisLockService
from unikit.contrib.redis.lock.dto import RedisLock
from unikit.lock.dto import Lock
from unikit.utils.async_utils import run_async


class AsyncRedisLockService(BaseRedisLockService):
    """Redis lock implementation based on async Redis client."""

    def __init__(self, redis_url: str) -> None:
        super().__init__()
        self.__redis = StrictRedis.from_url(redis_url)
        self.__skip_files = self._get_own_files_list() + (__file__,)

    async def aacquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> Lock | None:
        """Acquire a lock asynchronously."""
        lock = self._build_new_lock(op_name, target, op_params, timeout)
        return await self.__acquire_lock(lock)

    async def aget_lock(
        self, op_name: str | None = None, target: str | None = None, op_params: dict[str, Any] | None = None
    ) -> Lock | None:
        """Get a lock asynchronously."""
        full_op_name = self._create_op_name(op_name, op_params)
        target_id = self._create_target_id(target)
        lock_id = self._create_pk_for_lock(full_op_name, target_id)
        return await self.__get_lock(lock_id, keep_secret=False)

    async def arelease(self, lock: Lock) -> None:
        """Release a lock asynchronously."""
        existing_lock = await self.__get_lock(lock_id=lock.get_lock_id(), keep_secret=True)
        if existing_lock:
            if existing_lock.get_secret() != lock.get_secret():
                raise ValueError("Only lock owner can release the lock")
            await self.__redis.delete(lock.get_lock_id())

    async def aclean_all_locks(self) -> None:
        """Clean all locks asynchronously."""
        await self.__redis.flushdb()

    def acquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> Lock | None:
        """Acquire a lock synchronously."""
        return run_async(self.aacquire(op_name=op_name, target=target, op_params=op_params, timeout=timeout))

    def get_lock(
        self, op_name: str | None = None, target: str | None = None, op_params: dict[str, Any] | None = None
    ) -> Lock | None:
        """Get a lock synchronously."""
        return run_async(self.aget_lock(op_name=op_name, target=target, op_params=op_params))

    def release(self, lock: Lock) -> None:
        """Release a lock synchronously."""
        run_async(self.arelease(lock))

    def clean_all_locks(self) -> None:
        """Clean all locks synchronously."""
        run_async(self.aclean_all_locks())

    async def dispose(self) -> None:
        """Dispose the service."""
        await self.__redis.aclose()

    async def __acquire_lock(self, lock: RedisLock) -> RedisLock | None:
        acquired = await self.__redis.set(
            lock.get_lock_id(), value=json.dumps(lock.to_dict()), nx=True, pxat=lock.get_ts_expires()
        )
        return lock if acquired else None

    async def __get_lock(self, lock_id: str, keep_secret: bool = False) -> RedisLock | None:
        return self._parse_lock(
            await self.__redis.get(lock_id),
            keep_secret=keep_secret,
        )



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
import logging
from typing import TYPE_CHECKING, Any, Generic, Self, TypeVar

from django.apps import AppConfig
from django.http import HttpRequest
from injector import Injector, Provider

from unikit.abstract import Abstract, AbstractMeta
from unikit.di import root_container
from unikit.security_context import BaseSecurityContext

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser

logger = logging.getLogger(__name__)


class BaseDiSupportedApp(AppConfig, abc.ABC):  # type: ignore[misc]
    """Base class for applications that support Dependency Injection."""

    default = False

    def __init__(self, app_name: str, app_module: Any) -> None:
        super().__init__(app_name, app_module)
        self._di_container = root_container

    def _install_di_modules(self) -> None:
        """Scan current application for DI Module declaration and install discovered modules into root container."""
        logger.debug("Discovering DI modules for app %s", self.name)
        target_package_name = ".".join(self.__module__.split(".")[:-1])
        self._di_container.autodiscover_modules(target_package_name)

    def ready(self) -> None:
        """Django callback invoked when application is ready to be used."""
        super().ready()
        # Import DI context if any
        self._install_di_modules()


TDjangoUser = TypeVar("TDjangoUser", bound="AbstractUser")


class BaseDjangoSecurityContext(BaseSecurityContext, Generic[TDjangoUser], Abstract, metaclass=AbstractMeta):
    """Base class for Django Security Context."""

    @abc.abstractmethod
    def get_current_user(self) -> TDjangoUser | None:
        """Get current user or None if context is anonymous."""
        pass

    def get_current_user_id(self) -> str:
        """
        Return serializable identifier of the current user.

        This identifier will be used to pass user identity between services e.g. via message broker.
        """
        user = self.get_current_user()
        return str(user.pk) if user else ""

    def get_principal_id(self) -> str | None:
        """Get principal ID."""
        return self.get_current_user_id()

    def __str__(self) -> str:
        return f"SecurityContext(user={self.get_current_user()})"


class BaseWebSecurityContext(BaseDjangoSecurityContext, metaclass=abc.ABCMeta):
    """Base class for Web Security Context."""

    @classmethod
    @abc.abstractmethod
    def from_http_request(cls, request: HttpRequest) -> Self:
        """Create Web Security Context from HTTP Request."""
        pass


class WebSecurityContextProvider(Provider[BaseWebSecurityContext]):
    """Provider for Web Security Context."""

    def __init__(self, context_cls: type[BaseWebSecurityContext]):
        super().__init__()
        self._context_cls = context_cls

    def get(self, injector: Injector) -> BaseWebSecurityContext:
        """Get Web Security Context."""
        request = injector.get(HttpRequest)
        if request is None:
            raise ValueError(
                "HttpRequest is not available in DI context. "
                "Have you installed `django_di` and "
                "added `unikit.contrib.django.di.middleware.InjectRequestMiddleware` middleware?"
            )
        return self._context_cls.from_http_request(request)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
from typing import Any, Generator, Generic, TypeVar, cast

import injector
from typing_extensions import override

from unikit.dto.interfaces import Nameable

T = TypeVar("T")
D = TypeVar("D")
K = TypeVar("K")
TNameable = TypeVar("TNameable", bound=Nameable)


class AbstractRegistry(Generic[K, T], abc.ABC):
    """Abstract base class for any registry that defines all required interface methods."""

    def __init__(self, injector_: injector.Inject[injector.Injector]) -> None:
        super().__init__()
        self._registry: dict[K, T] = {}
        self._injector = injector_

    def register(self, name: K, obj: T) -> None:
        """Register an object in the registry."""
        self._registry[name] = obj

    @abc.abstractmethod
    def get(self, key: K) -> T | None:
        """Get an object from the registry."""
        pass

    def get_or_raise(self, key: K) -> T:
        """Get an object from the registry or raise a KeyError if it doesn't exist."""
        obj = self.get(key)
        if obj is None:
            raise KeyError(f"Key {key} not found in registry {self.__class__.__name__}")
        return obj

    def get_builder(self, key: K) -> injector.AssistedBuilder[T] | None:
        """Get an object builder from the registry."""
        obj = self._registry.get(key)
        if isinstance(obj, type):
            return self._injector.get(injector.AssistedBuilder[obj])  # type: ignore
        return None

    def get_builder_or_raise(self, key: K) -> injector.AssistedBuilder[T]:
        """Get an object builder from the registry or raise a KeyError if it doesn't exist."""
        obj = self.get_builder(key)
        if obj is None:
            raise KeyError(f"Key {key} not found in registry {self.__class__.__name__}")
        return obj

    def exists(self, key: K) -> bool:
        """Check if a key exists in the registry."""
        return key in self._registry

    @abc.abstractmethod
    def all(self) -> list[T]:
        """Return all registered objects as a list."""
        pass

    @abc.abstractmethod
    def get_all(self) -> Generator[T, T, None]:
        """Return all registered objects as a generator."""
        pass


class Registry(AbstractRegistry[K, T | type[T]], Generic[K, T]):
    """Simple registry implementation that stores objects in a dictionary."""

    def get(self, key: K) -> T | None:
        """Get an object from the registry."""
        obj = self._registry.get(key)
        if isinstance(obj, type):
            return cast(T, self._injector.get(obj))
        return obj

    def get_or_raise(self, key: K) -> T:
        """Get an object from the registry or raise a KeyError if it doesn't exist."""
        return cast(T, super().get_or_raise(key))

    def get_builder(self, key: K) -> injector.AssistedBuilder[T] | None:  # type: ignore[override]
        """Get an object builder from the registry."""
        return cast(injector.AssistedBuilder[T] | None, super().get_builder(key))

    def get_builder_or_raise(self, key: K) -> injector.AssistedBuilder[T]:  # type: ignore[override]
        """Get an object builder from the registry or raise a KeyError if it doesn't exist."""
        return cast(injector.AssistedBuilder[T], super().get_builder_or_raise(key))

    def all(self) -> list[T | type[T]]:
        """
        Return all registered objects.

        IMPORTANT classes won't be instantiated.
        :return: list of all registered objects
        """
        return list(self._registry.values())

    def get_all(self) -> Generator[T, T, None]:  # type: ignore[override]
        """
        Return all registered objects as a generator.

        :return: all registered objects
        """
        for key in self._registry.keys():
            yield self.get_or_raise(key)


class FactoryRegistry(AbstractRegistry[K, type[T]], Generic[K, T]):
    """Registry object which allows to use registered items as a base types for creating new objects."""

    @override
    def get(self, key: K) -> type[T] | None:
        """Get an object class from the registry."""
        obj_type = self._registry.get(key)
        return obj_type

    def build(self, key: K, params: dict[str, Any] | None = None) -> T | None:
        """
        Build the instance of the object from the registry with the given arguments.

        :param key: key of the object in the registry
        :param params: arguments for the object creation.
            None means that `__init__` method doesn't require any arguments except injected ones or defaults.
        """
        obj_type = self.get(key)
        if obj_type is None:
            return None
        return self._injector.create_object(obj_type, additional_kwargs=params)

    def build_or_raise(self, key: K, params: dict[str, Any] | None = None) -> T:
        """
        Build the instance of the object from the registry with the given arguments.

        :param key: key of the object in the registry
        :param params: arguments for the object creation

        Raise a KeyError if the object doesn't exist.
        """
        obj = self.build(key, params)
        if obj is None:
            raise KeyError(f"Key {key} not found in registry {self.__class__.__name__}")
        return obj

    def all(self) -> list[type[T]]:
        """
        Return all registered objects.

        :return: list of all registered objects
        """
        return list(self._registry.values())

    def get_all(self) -> Generator[type[T], type[T], None]:
        """
        Return all registered objects as a generator.

        :return: all registered objects
        """
        for key in self._registry.keys():
            yield self.get_or_raise(key)

    def build_all(self, params: dict[str, Any] | None = None) -> Generator[T, T, None]:
        """
        Build all registered objects with the given arguments.

        :param params: arguments for the object creation
        :return: all registered objects
        """
        for key in self._registry.keys():
            yield self.build_or_raise(key, params)


class NamedRegistry(Registry[str, TNameable], Generic[TNameable]):
    """Registry which allows to register objects by name."""

    def get_names(self) -> set[str]:
        """Return names of all registered objects."""
        return set(self._registry.keys())


class NamedFactoryRegistry(FactoryRegistry[str, TNameable], Generic[TNameable]):
    """Registry which allows to register objects by name."""

    def get_names(self) -> set[str]:
        """Return names of all registered objects."""
        return set(self._registry.keys())


class DefaultAwareMixin(Generic[D], metaclass=abc.ABCMeta):
    """Mixing for registries which have a default value."""

    def __init__(self, default: D | None = None) -> None:
        super().__init__()
        self.__default: D | None = default

    def get_default(self) -> D | None:
        """Return the default value for this registry."""
        return self.__default

    def set_default(self, default: D) -> None:
        """Set the default value for this registry."""
        self.__default = default



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import dataclasses
from typing import Any, Mapping, TypeVar, cast

from pydantic import BaseModel

from unikit.utils.default import OnErrorDef, raise_or_default

TDict = TypeVar("TDict", bound=Mapping)
_T = TypeVar("_T")


def deepmerge(to_dict: TDict, from_dict: TDict, merge_lists: bool = False) -> TDict:
    """
    Merge two dictionaries recursively and return created dict. Dict `from_dict` will be merged into dict `to_dict`.

    This is a PURE function (doesn't have side effects), so it will not modify the original dictionaries.
    If a key is present in both dictionaries, the value from `new` will be used.

    :param to_dict: original object which will be used as a base
    :param from_dict: new object which will be merged into original
    :param merge_lists: whether to merge lists or replace them
    """

    def merge(_old: TDict, _new: TDict) -> TDict:
        # pylint: disable=no-else-return
        if isinstance(_new, dict):
            if not isinstance(_old, dict):
                return cast(TDict, _new)
            res = _old.copy()
            for k, v in _new.items():
                res[k] = merge(_old[k], v) if k in _old else v
            return cast(TDict, res)
        elif isinstance(_new, list):
            if merge_lists:
                if not isinstance(_old, list):
                    return _new
                return _old + _new
            else:
                return _new
        return _new

    return merge(to_dict, from_dict)


def set_objects(_dict: dict, *objects: Any, key: str | None = None) -> None:
    """
    Convert all given objects to dictionaries and set them to the target dictionary.

    If key is provided - all objects will be set under this key in the target dictionary.

    :param _dict: target dictionary
    :param objects: objects to be serialized and set to the target dictionary
    :param key: optional key to set objects into
    """
    result_dict: dict[str, Any] = {}
    for obj in objects:
        if dataclasses.is_dataclass(obj):
            obj_dict = dataclasses.asdict(obj)  # type: ignore[call-overload]
        elif hasattr(obj, "to_dict"):
            obj_dict = obj.to_dict()
        elif isinstance(obj, dict):
            obj_dict = obj
        elif isinstance(obj, BaseModel):
            obj_dict = obj.model_dump(by_alias=True)
        else:
            raise ValueError("Object must be either dict, dataclass or has .to_dict() method")
        result_dict.update(obj_dict)
    if key is None:
        _dict.update(result_dict)
    else:
        _dict[key] = result_dict


def get_object(
    _dict: TDict, target_cls: type[_T], key: str | None = None, on_missing: OnErrorDef[Any] = None
) -> _T | None:
    """
    Get the object from the dictionary and convert it to the target class.

    :param _dict: target dictionary
    :param target_cls: target class
    :param key: optional key, if set the data will be extracted from this specific key, otherwise from the dict itself.
    :param on_missing: what to do if the key is missing. Could be either error or default value.
    :return: instance of target class populated from the dict
    """
    target_object = _dict.get(key) if key else _dict
    if target_object is None:
        return raise_or_default(
            on_missing, f"Key `{key}` not found in the dictionary" if key else "Dictionary is empty"
        )
    if dataclasses.is_dataclass(target_cls):
        fields = set([f.name for f in dataclasses.fields(target_cls)])
        data: dict[str, Any] = {}
        for k, v in target_object.items():
            if k in fields:
                data[k] = v
        obj = target_cls(**data)
    elif issubclass(target_cls, BaseModel):
        obj = target_cls.model_load(target_object)  # type: ignore
    elif hasattr(target_cls, "from_dict"):
        obj = target_cls.from_dict(target_object)  # type: ignore
    else:
        raise ValueError("Object must be either dict, dataclass or has .from_dict() method")
    return cast(_T, obj)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from django.apps import apps
from django.http import HttpRequest
from django.utils.deprecation import MiddlewareMixin


class InjectRequestMiddleware(MiddlewareMixin):  # type: ignore[misc]
    """Middleware that injects the current request into the injector."""

    app = apps.get_app_config("django_di")

    def process_request(self, request: HttpRequest) -> None:
        """Set current request to the injector."""
        self.app.django_module.set_request(request)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import shutil
from typing import IO


def copy_stream(src_stream: IO, target_stream: IO, chunk_size: int = 4096) -> None:
    """
    Copy the contents of one stream to another.

    :param src_stream: the input stream to copy data from
    :param target_stream: the output stream to copy data to
    :param chunk_size: the size of the chunks that data will be read and written in. Defaults to 4096
    :return: None
    """
    shutil.copyfileobj(src_stream, target_stream, length=chunk_size)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import injector

from unikit.contrib.taskiq.di import BaseTaskiqDiModule
from unikit.worker import WorkerServiceRegistry


class _TaskiqModule(BaseTaskiqDiModule):

    def configure(self, binder: injector.Binder) -> None:
        super().configure(binder)

        self.register_singleton(WorkerServiceRegistry)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from functools import cached_property
from typing import TYPE_CHECKING

from django.apps import apps
from django.core.management import BaseCommand

if TYPE_CHECKING:
    from unikit.contrib.django.taskiq.apps import TaskiqConfig


class BaseTaskiqCommand(BaseCommand):  # type: ignore[misc]
    """Base class for Taskiq management commands."""

    @cached_property
    def taskiq_app(self) -> "TaskiqConfig":
        """Return Taskiq application configuration."""
        return apps.get_app_config("django_taskiq")

    def _get_broker_name(self, args: tuple[str, ...]) -> str:
        if len(args) > 0 and not args[0].startswith("-"):
            broker_name = args[0]
            args = args[1:]
        else:
            broker_name = self.taskiq_app.default_broker_name
        return broker_name

    @staticmethod
    def _to_taskiq_class_path(path: str) -> str:
        return ":".join((path.rsplit(".", 1)))



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import datetime
from typing import Any, Generic

from taskiq import TaskiqMessage, TaskiqMiddleware, TaskiqResult
from taskiq.depends.progress_tracker import TaskProgress, TaskState

from unikit.contrib.taskiq.di import default_current_task_holder
from unikit.contrib.taskiq.dto import (
    TASKIQ_LABEL_DATE_POSTED,
    TASKIQ_LABEL_SECURITY_CONTEXT,
    TASKIQ_LABEL_TASKNAME,
    TaskiqTask,
)
from unikit.di import root_container
from unikit.security_context import (
    ContextVarSecurityContextHolder,
    SecurityContextDto,
    TBaseSecurityContext,
    default_security_context_holder,
)
from unikit.utils.logger import LogMixin
from unikit.utils.time_utils import datetime_now


class TaskInfoLabelsMiddleware(TaskiqMiddleware):
    """Middleware to attach task name, date posted and other metadata to labels."""

    async def pre_send(self, message: TaskiqMessage) -> TaskiqMessage:
        """Pre-send hook to attach task name to Taskiq message."""
        message.labels[TASKIQ_LABEL_TASKNAME] = message.task_name
        message.labels[TASKIQ_LABEL_DATE_POSTED] = datetime_now().isoformat()
        return message


class TaskInfoDiMiddleware(TaskiqMiddleware, LogMixin):
    """
    Middleware to put current task information into DI context.

    This allows user to inject `TaskiqPostedTask` or `PostedTask` into their services.
    """

    async def pre_execute(self, message: TaskiqMessage) -> TaskiqMessage:
        """Pre-execute hook to TaskiqTask available in DI context."""
        time_posted_str = message.labels.get(TASKIQ_LABEL_DATE_POSTED)
        time_posted = datetime.datetime.fromisoformat(time_posted_str) if time_posted_str else None
        if time_posted is None:
            self.log.error("Task doesn't have a date posted label. Have you forgot to add TaskInfoLabelsMiddleware?")
        default_current_task_holder.set_current_task(
            TaskiqTask(
                uuid=message.task_id,
                timestamp=time_posted or datetime_now(),
                task_name=message.task_name,
                message=message,
            )
        )
        return message

    async def post_execute(self, message: TaskiqMessage, result: TaskiqResult[Any]) -> None:
        """Post-execute hook to clean up task info."""
        default_current_task_holder.set_current_task(None)

    async def on_error(self, message: TaskiqMessage, result: TaskiqResult[Any], exception: BaseException) -> None:
        """On error hook to clean up task info."""
        default_current_task_holder.set_current_task(None)


class ReportExecutionStartedMiddleware(TaskiqMiddleware):
    """Middleware to report execution start time."""

    def __init__(self, include_labels: bool = True) -> None:
        super().__init__()
        self.include_labels = include_labels

    async def pre_execute(self, message: TaskiqMessage) -> TaskiqMessage:
        """Pre-execute hook to attach execution start time to Taskiq message."""
        metadata = {}
        if self.include_labels:
            metadata["labels"] = message.labels
        await self.broker.result_backend.set_progress(
            message.task_id, TaskProgress(state=TaskState.STARTED, meta=metadata)
        )
        return message


class SecurityContextMiddleware(TaskiqMiddleware, Generic[TBaseSecurityContext]):
    """Middleware to attach security context to Taskiq messages."""

    def __init__(
        self,
        security_context_interface_cls: type[TBaseSecurityContext],
        security_context_holder: ContextVarSecurityContextHolder | None = None,
    ) -> None:
        self._security_context_interface_cls: type[TBaseSecurityContext] = security_context_interface_cls
        self._security_context_holder = security_context_holder or default_security_context_holder

    async def pre_send(self, message: TaskiqMessage) -> TaskiqMessage:
        """Pre-send hook to attach security context properties to Taskiq message."""
        try:
            security_context = root_container.get(self._security_context_interface_cls)
        except Exception:
            raise Exception(
                "Security context is not available in DI container. Make sure you've registered it "
                "like it is shown in the docs for ContextVarSecurityContextHolder."
            )
        await self._attach_context_to_message(message, security_context)
        return message

    async def pre_execute(self, message: TaskiqMessage) -> TaskiqMessage:
        """Pre-execute hook to create security context and make it available in DI context."""
        security_context = await self._create_security_context(message)
        default_security_context_holder.set_security_context(security_context)
        return message

    async def post_execute(self, message: TaskiqMessage, result: TaskiqResult[Any]) -> None:
        """Post-execute hook to clean up security context."""
        default_security_context_holder.set_security_context(None)

    async def on_error(self, message: TaskiqMessage, result: TaskiqResult[Any], exception: BaseException) -> None:
        """On error hook to clean up security context."""
        default_security_context_holder.set_security_context(None)

    async def _create_security_context(self, message: TaskiqMessage) -> TBaseSecurityContext:
        return await self._security_context_interface_cls.afrom_dto(
            SecurityContextDto(message.labels.get(TASKIQ_LABEL_SECURITY_CONTEXT, {}))
        )

    async def _attach_context_to_message(
        self, message: TaskiqMessage, security_context: TBaseSecurityContext
    ) -> TaskiqMessage:
        message.labels[TASKIQ_LABEL_SECURITY_CONTEXT] = security_context.to_dto()
        return message



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import datetime

from unikit.utils.type_utils import TAnyDate


def datetime_now() -> datetime.datetime:
    """Return `now` in UTC."""
    return datetime.datetime.now(tz=datetime.timezone.utc)


def get_midnight(date: datetime.datetime | None = None) -> datetime.datetime:
    """Return datetime pointing to the midnight of the given date or today."""
    return (date or datetime_now()).replace(hour=0, minute=0, second=0, microsecond=0)


def last_day_of_month(date_: TAnyDate) -> TAnyDate:
    """Return the last day of the month for the given date."""
    if date_.month == 12:
        return date_.replace(day=31)
    return date_.replace(month=date_.month + 1, day=1) - datetime.timedelta(days=1)


def first_day_of_month(d: TAnyDate) -> TAnyDate:
    """Return the first day of the month for the given date."""
    return d.replace(day=1)


def first_day_of_next_month(date: TAnyDate) -> TAnyDate:
    """Return the first day of the next month for the given date."""
    return last_day_of_month(date) + datetime.timedelta(days=1)


def get_month_boundaries(month: datetime.datetime) -> tuple[datetime.datetime, datetime.datetime]:
    """Return tuple of the earliest and the latest passed month datetimes."""
    earliest = first_day_of_month(month).replace(hour=0, minute=0, second=0, microsecond=0)
    latest = last_day_of_month(month).replace(hour=23, minute=59, second=59, microsecond=999999)
    return earliest, latest



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import injector

from unikit.contrib.taskiq.dto import TaskiqTask
from unikit.di import DiModule, DiProxy
from unikit.worker import ContextVarWorkerTaskHolder, ContextVarWorkerTaskProvider, PostedTask


class BaseTaskiqDiModule(DiModule):
    """Base DI module for Taskiq."""

    def configure(self, binder: injector.Binder) -> None:
        """Configure DI module."""
        super().configure(binder)

        provider = injector.InstanceProvider(
            PostedTaskProvider(
                ContextVarWorkerTaskProvider(default_current_task_holder), binder.injector  # type: ignore
            )
        )
        binder.bind(PostedTask, provider)
        binder.bind(TaskiqTask, provider)


default_current_task_holder = ContextVarWorkerTaskHolder()


class PostedTaskProvider(DiProxy[TaskiqTask]):
    """Provider for TaskiqPostedTask."""

    pass



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc

from django.db import models


class AbstractModelMeta(abc.ABCMeta, type(models.Model)):  # type: ignore[misc]
    """
    Metaclass for abstract models that allows to inherit from both abstract models and models.Model.

    Usage:

    class MyAbstractModel(SomeOtherModel, models.Model, metaclass=AbstractModelMeta):
        ...

        @abc.abstractmethod
        def do_something(self):
            pass
    """

    pass



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from contextlib import contextmanager
import logging
from typing import Any, ContextManager, Generator


class StaticFieldFilter(logging.Filter):
    """Filter to add static fields to log records."""

    def __init__(self, fields: dict, prefix: str = ""):
        super().__init__()
        self.static_fields: dict = fields
        self._prefix = prefix

    def filter(self, record: logging.LogRecord) -> bool:
        """Add static fields to log record."""
        for k, v in self.static_fields.items():
            setattr(record, self._prefix + k, v)
        return True


class LogMixin:
    """Convenience super-class to have a logger configured with the class name."""

    _log: logging.Logger | None = None

    @property
    def log(self) -> logging.Logger:
        """Return a logger."""
        if self._log is None:
            self._log = logging.getLogger(self.__class__.__module__ + "." + self.__class__.__name__)
        return self._log

    def with_logging_context(self, **extras: Any) -> ContextManager[None]:
        """
        Context manager that adds extra fields to log records.

        :param extras: any extra fields to be added to log every record
        """
        return with_logging_context(self.log, **extras)


@contextmanager
def with_logging_context(logger: logging.Logger, prefix: str = "", **extras: Any) -> Generator[Any, None, None]:
    """
    Context manager that adds extra fields to log records.

    :param logger: logger instance to be modified
    :param prefix: prefix to be added to every key in extras
    :param extras: any extra fields to be added to log every record
    """
    logger_filter = StaticFieldFilter(extras, prefix)
    try:
        logger.addFilter(logger_filter)
        yield
    finally:
        logger.removeFilter(logger_filter)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import dataclasses

from taskiq import AsyncTaskiqTask, TaskiqMessage

from unikit.worker import PostedTask


@dataclasses.dataclass(kw_only=True, frozen=True)
class TaskiqPostedTask(PostedTask):
    """TaskiqPostedTask is a DTO for Taskiq task."""

    task: AsyncTaskiqTask


@dataclasses.dataclass(kw_only=True, frozen=True)
class TaskiqTask(PostedTask):
    """TaskiqTask is a DTO for Taskiq task."""

    message: TaskiqMessage


TASKIQ_LABEL_SECURITY_CONTEXT = "unikit_security_context"
TASKIQ_LABEL_TASKNAME = "__taskiq__task_name"
TASKIQ_LABEL_DATE_POSTED = "__taskiq__date_posted"



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import datetime
from typing import Any

from asgiref.sync import sync_to_async

from unikit.lock.dto import Lock, SimpleLockDto
from unikit.lock.service import LockService


class NoopLockService(LockService):
    """A lock service that does nothing."""

    def acquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> Lock | None:
        """Acquire a lock synchronously."""
        timeout = timeout or datetime.timedelta(minutes=5)
        assert op_name or target, "Either op_name or target must be specified"
        target_id = self._create_target_id(target)
        full_op_name = self._create_op_name(op_name, op_params)
        lock_id = self._create_pk_for_lock(full_op_name, target_id)
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        lock = SimpleLockDto(
            lock_id=lock_id,
            operation_name=op_name,
            target_id=target_id,
            ts_acquired=now,
            ts_expires=now + timeout,
            secret="",
            owner=None,
        )
        return lock

    def get_lock(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
    ) -> Lock | None:
        """Get a lock synchronously."""
        return None

    def release(self, lock: Lock) -> None:
        """Release a lock synchronously."""
        pass  # noop

    async def aacquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> Lock | None:
        """Acquire a lock asynchronously."""
        return await sync_to_async(self.acquire)(op_name, target, op_params, timeout)

    async def aget_lock(
        self, op_name: str | None = None, target: str | None = None, op_params: dict[str, Any] | None = None
    ) -> Lock | None:
        """Get a lock asynchronously."""
        return None

    async def arelease(self, lock: Lock) -> None:
        """Release a lock asynchronously."""
        pass

    def clean_all_locks(self) -> None:
        """Clean all locks synchronously."""
        return None

    async def aclean_all_locks(self) -> None:
        """Clean all locks asynchronously."""
        return None



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import datetime
import json
from typing import Any, Sequence

from asgiref.sync import sync_to_async
from redis import StrictRedis

from unikit.lock.dto import Lock

from .base import BaseRedisLockService
from .dto import RedisLock


class RedisLockService(BaseRedisLockService):
    """Redis lock implementation based on sync Redis client."""

    def __init__(self, redis_url: str) -> None:
        super().__init__()
        self.__redis = StrictRedis.from_url(redis_url)

    def acquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> Lock | None:
        """Acquire a lock synchronously."""
        lock = self._build_new_lock(op_name, target, op_params, timeout)
        return self.__acquire_lock(lock)

    def release(self, lock: Lock) -> None:
        """Release a lock synchronously."""
        existing_lock = self.__get_lock(lock_id=lock.get_lock_id(), keep_secret=True)
        if existing_lock:
            if existing_lock.get_secret() != lock.get_secret():
                raise ValueError("Only lock owner can release the lock")
            self.__redis.delete(lock.get_lock_id())

    def delete(self, lock_id: str) -> None:
        """Delete a lock synchronously."""
        self.__redis.delete(lock_id)

    def get_all(self) -> Sequence[Lock]:
        """Get all locks synchronously."""
        result: list[Lock] = []
        keys = self.__redis.keys("*")
        for k in keys:  # type: ignore[union-attr]
            lock = self.__get_lock(str(k, encoding="utf-8"))
            if lock:
                result.append(lock)
        return result

    def clean_all_locks(self) -> None:
        """Clean all locks synchronously."""
        self.__redis.flushdb()

    def get_lock(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
    ) -> Lock | None:
        """Get a lock synchronously."""
        full_op_name = self._create_op_name(op_name, op_params)
        target_id = self._create_target_id(target)
        lock_id = self._create_pk_for_lock(full_op_name, target_id)
        return self.__get_lock(lock_id, keep_secret=False)

    def extend_lock(self, lock: RedisLock, new_expiration_ts: datetime.datetime) -> Lock:
        """Extend a lock synchronously."""
        lock_id = lock.get_lock_id()
        existing_lock = self._parse_lock(self.__redis.get(lock_id), keep_secret=True)  # type: ignore
        if existing_lock and existing_lock.get_secret() == lock.get_secret():
            self.__redis.pexpireat(lock_id, new_expiration_ts)
            return lock.clone(ts_expires=new_expiration_ts)
        else:
            raise ValueError(
                f"Unable to extend lock {lock_id}. Lock doesn't exist, already expired or you do "
                f"not have permissions"
            )

    async def aacquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> Lock | None:
        """Acquire a lock asynchronously."""
        return await sync_to_async(self.acquire)(op_name=op_name, target=target, op_params=op_params, timeout=timeout)

    async def aget_lock(
        self, op_name: str | None = None, target: str | None = None, op_params: dict[str, Any] | None = None
    ) -> Lock | None:
        """Get a lock asynchronously."""
        return await sync_to_async(self.get_lock)(op_name=op_name, target=target, op_params=op_params)

    async def arelease(self, lock: Lock) -> None:
        """Release a lock asynchronously."""
        return await sync_to_async(self.release)(lock)

    async def aclean_all_locks(self) -> None:
        """Clean all locks asynchronously."""
        return await sync_to_async(self.clean_all_locks)()

    def __get_lock(self, lock_id: str, keep_secret: bool = False) -> RedisLock | None:
        return self._parse_lock(
            self.__redis.get(lock_id),  # type: ignore[arg-type]
            keep_secret=keep_secret,
        )

    def __acquire_lock(self, lock: RedisLock) -> RedisLock | None:
        acquired = self.__redis.set(
            lock.get_lock_id(), value=json.dumps(lock.to_dict()), nx=True, pxat=lock.get_ts_expires()
        )
        return lock if acquired else None

    @classmethod
    def _get_own_files_list(cls) -> tuple[str, ...]:
        return super()._get_own_files_list() + (__file__,)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
import dataclasses
import datetime

from unikit.abstract import Abstract, AbstractMeta


class Lock(Abstract, metaclass=AbstractMeta):
    """Lock interface."""

    @abc.abstractmethod
    def get_lock_id(self) -> str:
        """Get the lock ID."""
        pass

    @abc.abstractmethod
    def get_operation_name(self) -> str | None:
        """Get the operation name."""
        pass

    @abc.abstractmethod
    def get_target_id(self) -> str | None:
        """Get the target ID."""
        pass

    @abc.abstractmethod
    def get_secret(self) -> str:
        """Get the lock secret."""
        pass

    @abc.abstractmethod
    def get_owner(self) -> str | None:
        """Get the lock owner."""
        pass


@dataclasses.dataclass(kw_only=True)
class SimpleLockDto(Lock):
    """Simple lock data transfer object."""

    lock_id: str
    operation_name: str | None
    target_id: str | None
    secret: str
    owner: str | None
    ts_acquired: datetime.datetime
    ts_expires: datetime.datetime

    def get_lock_id(self) -> str:
        """Get the lock ID."""
        return self.lock_id

    def get_operation_name(self) -> str | None:
        """Get the operation name."""
        return self.operation_name

    def get_target_id(self) -> str | None:
        """Get the target ID."""
        return self.target_id

    def get_secret(self) -> str:
        """Get the lock secret."""
        return self.secret

    def get_owner(self) -> str | None:
        """Get the lock owner."""
        return self.owner



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
from contextlib import asynccontextmanager, contextmanager
import datetime
import time
from types import FrameType
from typing import Any, AsyncIterator, Iterator, Never

from unikit.abstract import Abstract, AbstractMeta
from unikit.lock.dto import Lock


class LockService(Abstract, metaclass=AbstractMeta):
    """Abstract base class for lock services."""

    _MAX_CALL_STACK_DEPTH = 5

    @abc.abstractmethod
    def acquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> Lock | None:
        """
        Acquire a lock synchronously.

        :param op_name: The name of the operation.
        :param target: The target for the lock.
        :param op_params: Additional parameters for the operation.
        :param timeout: The timeout for the lock.
        :return: The acquired lock or None if the lock could not be acquired.
        """
        pass

    @abc.abstractmethod
    async def aacquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> Lock | None:
        """
        Acquire a lock asynchronously.

        :param op_name: The name of the operation.
        :param target: The target for the lock.
        :param op_params: Additional parameters for the operation.
        :param timeout: The timeout for the lock.
        :return: The acquired lock or None if the lock could not be acquired.
        """
        pass

    @abc.abstractmethod
    def get_lock(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
    ) -> Lock | None:
        """
        Retrieve an existing lock synchronously.

        :param op_name: The name of the operation.
        :param target: The target for the lock.
        :param op_params: Additional parameters for the operation.
        :return: The retrieved lock or None if the lock does not exist.
        """
        pass

    @abc.abstractmethod
    async def aget_lock(
        self,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
    ) -> Lock | None:
        """
        Retrieve an existing lock asynchronously.

        :param op_name: The name of the operation.
        :param target: The target for the lock.
        :param op_params: Additional parameters for the operation.
        :return: The retrieved lock or None if the lock does not exist.
        """
        pass

    @contextmanager
    def wait_and_acquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        waiting_timeout: datetime.timedelta | None = None,
        timeout: datetime.timedelta | None = None,
        op_params: dict[str, Any] | None = None,
    ) -> Iterator[Never]:
        """
        Context manager to wait and acquire a lock synchronously.

        :param op_name: The name of the operation.
        :param target: The target for the lock.
        :param waiting_timeout: The maximum time to wait for acquiring the lock.
        :param timeout: The timeout for the lock.
        :param op_params: Additional parameters for the operation.
        """
        timeout = timeout or datetime.timedelta(minutes=5)
        start = datetime.datetime.now()
        lock = self.acquire(op_name, target, timeout=timeout, op_params=op_params)
        while not lock:
            if waiting_timeout and datetime.datetime.now() - start > waiting_timeout:
                lock = self.get_lock(op_name, target, op_params)
                owner = lock.get_owner() if lock else None
                raise TimeoutError(
                    f"Cannot acquire lock(op={op_name}, target={target or 'n/a'}) within `{waiting_timeout}` timeout. "
                    f"Lock owner: {owner or 'n/a'}"
                )
            time.sleep(0.2)
            lock = self.acquire(op_name, target, op_params=op_params, timeout=timeout)
        try:
            yield  # type: ignore[misc]
        finally:
            self.release(lock)

    @contextmanager
    def with_lock(self, lock: Lock | None) -> Iterator[Never]:
        """Context manager which will release a lock (if any given) at the end."""
        try:
            yield  # type: ignore[misc]
        finally:
            if lock:
                self.release(lock)

    @asynccontextmanager
    async def awith_lock(self, lock: Lock | None) -> AsyncIterator[Never]:
        """Context manager which will release a lock (if any given) at the end."""
        try:
            yield  # type: ignore[misc]
        finally:
            if lock:
                await self.arelease(lock)

    @asynccontextmanager
    async def await_and_acquire(
        self,
        op_name: str | None = None,
        target: str | None = None,
        waiting_timeout: datetime.timedelta | None = None,
        timeout: datetime.timedelta | None = None,
        op_params: dict[str, Any] | None = None,
    ) -> AsyncIterator[Never]:
        """
        Async context manager to wait and acquire a lock asynchronously.

        :param op_name: The name of the operation.
        :param target: The target for the lock.
        :param waiting_timeout: The maximum time to wait for acquiring the lock.
        :param timeout: The timeout for the lock.
        :param op_params: Additional parameters for the operation.
        """
        timeout = timeout or datetime.timedelta(minutes=5)
        start = datetime.datetime.now()
        lock = await self.aacquire(op_name, target, timeout=timeout, op_params=op_params)
        while not lock:
            if waiting_timeout and datetime.datetime.now() - start > waiting_timeout:
                lock = await self.aget_lock(op_name, target, op_params)
                owner = lock.get_owner() if lock else None
                raise TimeoutError(
                    f"Cannot acquire lock(op={op_name}, target={target or 'n/a'}) within `{waiting_timeout}` timeout. "
                    f"Lock owner: {owner or 'n/a'}"
                )
            time.sleep(0.2)
            lock = await self.aacquire(op_name, target, op_params=op_params, timeout=timeout)
        try:
            yield  # type: ignore[misc]
        finally:
            await self.arelease(lock)

    @abc.abstractmethod
    def release(self, lock: Lock) -> None:
        """
        Release a lock synchronously.

        :param lock: The lock to be released.
        """
        pass

    @abc.abstractmethod
    async def arelease(self, lock: Lock) -> None:
        """
        Release a lock asynchronously.

        :param lock: The lock to be released.
        """
        pass

    @abc.abstractmethod
    def clean_all_locks(self) -> None:
        """Clean all locks synchronously."""
        pass

    @abc.abstractmethod
    async def aclean_all_locks(self) -> None:
        """Clean all locks asynchronously."""
        pass

    @classmethod
    def _create_target_id(cls, target: str | None = None) -> str | None:
        tpl = "node__{}"
        if target is None:
            return None
        elif isinstance(target, str):
            return tpl.format(target)
        raise ValueError(f"Unsupported target type: {type(target)}")

    @classmethod
    def _create_op_name(cls, op_name: str | None, op_params: dict[str, Any] | None) -> str | None:
        if not op_name:
            return None
        if not op_params:
            return op_name
        param_list = [f"{param}__{val}" for param, val in op_params.items()]
        return op_name + "|" + "|".join(param_list)

    @classmethod
    def _create_pk_for_lock(cls, op_name: str | None, target_id: str | None) -> str:
        if target_id:
            return f"target:{target_id}"
        elif op_name:
            return f"op:{op_name}"
        raise ValueError("Either name or target must be specified")

    @classmethod
    def _get_own_files_list(cls) -> tuple[str, ...]:
        """Return a list of files which needs to be dropped while scanning call-stack to resolve lock owner."""
        return __file__, "contextlib.py"

    @classmethod
    def _get_lock_owner(cls, frame: FrameType | None, skip_files: tuple[str, ...]) -> str | None:
        """
        Scan call stack to find the caller of `acquire` method outside LockService.

        It will be considered the owner of the lock.

        :param frame: call-stack frame to start analysis from
        :return: string describing the owner of the lock
        """
        depth = 0

        while frame and depth < cls._MAX_CALL_STACK_DEPTH:
            if frame.f_code.co_filename:
                should_skip = False
                for x in skip_files:
                    if frame.f_code.co_filename.endswith(x):
                        should_skip = True
                        break
                if not should_skip:
                    return "{} at {}:{}".format(
                        frame.f_code.co_name, frame.f_code.co_filename, frame.f_code.co_firstlineno
                    )
            frame = frame.f_back
            depth += 1
        return None



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
from typing import Any, Generic

from celery import Task
from celery._state import current_task, pop_current_task, push_current_task
import injector

from unikit.utils.type_utils import T


class TaskBodyMeta(abc.ABCMeta):
    """
    Metaclass for TaskBody classes that makes class compatible with Celery's shared_task/task decorator.

    See TaskBody for more information.
    """

    def __new__(mcs, name: str, bases: tuple[type, ...], namespace: dict[str, Any], **kwargs: Any) -> type:
        """Make class compatible with Celery's shared_task/task decorator."""
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        if not hasattr(cls, "run"):
            raise ValueError(f"TaskBodyMeta meta must be used with class defining run method. Was used with {name}")
        cls.__call__ = cls.run  # type: ignore[method-assign]
        return cls


class TaskBody(metaclass=TaskBodyMeta):
    """Base class for class based task definitions. Compatible only with TaskWithInjection Celery tasks."""

    @abc.abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> Any:
        """Implement in child classes to define task logic."""
        pass


class TaskWithInjection(Task, abc.ABC):  # type: ignore[misc]
    """A Celery task that can use Inject annotations to resolve dependencies prior invocation."""

    def _get_di_container(self) -> injector.Injector:
        container: injector.Injector | None = getattr(self, "di_container", None)
        if not container:
            container = getattr(self.app, "di_container", None)
        if not container:
            raise RuntimeError("DI Container unavailable in Celery context")
        return container

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Inject dependencies and call the run method."""
        push_current_task(self)
        if self.request:
            self.request_stack.push(self.request)
        try:
            if isinstance(self.run, type) and issubclass(self.run, TaskBody):
                runnable = self._get_di_container().create_object(self.run)
                return self._get_di_container().call_with_injection(runnable.run, args=args, kwargs=kwargs)
            else:
                return self._get_di_container().call_with_injection(self.run, args=args, kwargs=kwargs)
        finally:
            self.pop_request()
            pop_current_task()


class TaskRelatedProvider(injector.Provider[T], Generic[T]):
    """
    Resolves the target class from the current task's DI container if it exists.

    This might be useful if you want to define a child DI container for every task with some task-specific dependencies
    and resolve them in the task body.
    To use this provider you have to set DI container to the task instance in the task_prerun signal handler.
    Example:
        @task_prerun.connect
        def my_task_prerun_handler(task_id: str, task: MyBaseTask, kwargs: dict[str, Any], **other: Any) -> None:
            task.di_container = root_container.create_child_injector()
    """

    def __init__(self, target_cls: type[T]) -> None:
        self.target_cls = target_cls

    def get(self, injector_: injector.Injector) -> T:
        """Resolve the target class from the current task's DI container if it exists."""
        if current_task and hasattr(current_task, "di_container"):
            return current_task.di_container.get(self.target_cls)
        raise injector.UnsatisfiedRequirement(None, interface=self.target_cls)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from typing import Any

from taskiq.cli.worker.args import WorkerArgs
from taskiq.cli.worker.run import start_listen

from unikit.contrib.django.taskiq.management.base import BaseTaskiqCommand


class Command(BaseTaskiqCommand):
    """Start Taskiq worker process."""

    help = "Starts Taskiq worker process"

    def handle(self, *args: Any, **options: Any) -> None:
        """Start Taskiq worker process."""
        taskiq_app = self.taskiq_app
        broker_name = self._get_broker_name(args)

        broker_path_for_taskiq = self._to_taskiq_class_path(taskiq_app.broker_paths[broker_name])
        args = tuple([broker_path_for_taskiq] + list(args))
        worker_args = WorkerArgs.from_cli(args)
        worker_args.tasks_pattern = taskiq_app.task_discovery_pattern

        if not worker_args.modules:
            worker_args.fs_discover = True

        start_listen(worker_args)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
import dataclasses
import datetime
from typing import Generic, Self, TypeVar

from unikit.utils.logger import LogMixin


@dataclasses.dataclass(kw_only=True)
class ProgressState:
    """Dto for progress state."""

    pct: float | None = None
    items_done: int | None = None
    items_success: int | None = None
    items_failed: int | None = None
    items_skipped: int | None = None
    items_total: int | None = None
    eta_seconds: int | None = None
    failed_items: dict[str, str] = dataclasses.field(default_factory=dict)

    @property
    def progress_percent(self) -> float | None:
        """Get the progress percentage."""
        if self.pct is not None:
            return self.pct
        elif self.items_done and self.items_total:
            return (self.items_done / self.items_total) * 100.0
        return None

    @property
    def eta(self) -> datetime.timedelta | None:
        """Get the estimated time of completion."""
        if self.eta_seconds is not None:
            return datetime.timedelta(seconds=self.eta_seconds)
        return None

    def register_successful(self, num: int = 1) -> Self:
        """Register successful item."""
        if self.items_done is None:
            self.items_done = 0
        self.items_done += num
        if self.items_success is None:
            self.items_success = 0
        self.items_success += num
        return self

    def register_failed(self, num: int = 1, item_id: str | None = None, fail_reason: str | None = None) -> Self:
        """Register failed item."""
        if self.items_done is None:
            self.items_done = 0
        if self.items_failed is None:
            self.items_failed = 0
        self.items_failed += num
        if item_id:
            self.failed_items[item_id] = fail_reason or ""
        return self

    def register_failed_with_details(self, item_id: str, fail_reason: str | None = None) -> Self:
        """Register failed item with details."""
        if self.items_done is None:
            self.items_done = 0
        self.items_done += 1
        if self.items_failed is None:
            self.items_failed = 0
        self.items_failed += 1
        if item_id:
            self.failed_items[item_id] = fail_reason or ""
        return self

    def register_skipped(self, num: int = 1) -> Self:
        """Register skipped item."""
        if self.items_done is None:
            self.items_done = 0
        self.items_done += num
        if self.items_skipped is None:
            self.items_skipped = 0
        self.items_skipped += num
        return self

    def register_progress(
        self, success: int | None = None, failed: int | None = None, skipped: int | None = None
    ) -> Self:
        """Register progress."""
        self.register_successful(success or 0).register_failed(failed or 0).register_skipped(skipped or 0)
        return self

    def register_percent(self, pct: float) -> Self:
        """Register progress percentage."""
        self.pct = pct
        self.items_done = None
        self.items_total = None
        self.items_failed = None
        self.items_skipped = None
        self.items_success = None
        return self

    def reset_batch_size(self, total_items: int) -> Self:
        """
        Set total items count to 0.

        This also enables percentage tracking based on processed items.
        """
        self.items_total = total_items
        self.items_failed = self.items_done = self.items_skipped = self.items_success = 0
        return self

    @property
    def is_empty(self) -> bool:
        """Return True if the progress state is empty."""
        return self.items_total is None and self.pct is None

    @classmethod
    def create_by_items(cls, total_items: int, eta: datetime.timedelta | None = None) -> Self:
        """Create progress state by total items count."""
        return cls(items_total=total_items, items_done=0, eta_seconds=int(eta.total_seconds()) if eta else None)


TProgressState = TypeVar("TProgressState", bound=ProgressState)


class ProgressTracker(Generic[TProgressState], LogMixin, metaclass=abc.ABCMeta):
    """
    Progress tracker which allows long bulk processing tasks to report execution progress.

    This object establishes abstraction layer between task and result backend which used to capture progress and result.
    Typical usage:

    1. Your task creates an instance of tracker.
    2. You pass an instance as a parameter to your service which does some bulk processing.
    3. Service updates status like `tracker.register_success()`
    4. Service tracks progress by calling `tracker.track()` or `tracker.atrack()`
    """

    def __init__(self, progress_state: TProgressState, report_every_x_updates: int = 5):
        self._state = progress_state
        self.report_every_x_updates = report_every_x_updates
        self._untracked_updates = 0

    @property
    def state(self) -> TProgressState:
        """Get the current progress state."""
        return self._state

    def track(self) -> None:
        """
        Register progress update.

        This method should be called by the service to report progress updates. Under the hood it invokes storage
        backend to persist progres data.
        """
        self._untracked_updates += 1
        try:
            if self._untracked_updates >= self.report_every_x_updates:
                self._do_report_update()
        finally:
            self._untracked_updates = 0

    async def atrack(self) -> None:
        """
        Register progress update.

        This method should be called by the service to report progress updates. Under the hood it invokes storage
        backend to persist progres data.
        """
        self._untracked_updates += 1
        try:
            if self._untracked_updates >= self.report_every_x_updates:
                await self._ado_report_update()
        finally:
            self._untracked_updates = 0

    async def init(self) -> None:
        """
        Initialize the asynchronous components if needed.

        Should be overriden by child class.
        """
        pass

    @abc.abstractmethod
    def fetch_state(self) -> TProgressState:
        """Fetch state from the backend."""
        pass

    @abc.abstractmethod
    async def afetch_state(self) -> TProgressState:
        """Fetch state from the backend."""
        pass

    @abc.abstractmethod
    def _do_report_update(self) -> None:
        pass

    @abc.abstractmethod
    async def _ado_report_update(self) -> None:
        pass


class SimpleProgressTracker(ProgressTracker[TProgressState], Generic[TProgressState]):
    """Simple progress tracker which does not publish progres, just allows tracking internal state."""

    def fetch_state(self) -> TProgressState:
        """Noop."""
        return self.state

    async def afetch_state(self) -> TProgressState:
        """Noop."""
        return self.state

    def _do_report_update(self) -> None:
        pass

    async def _ado_report_update(self) -> None:
        pass

    @classmethod
    def create(cls) -> "SimpleProgressTracker[ProgressState]":
        """Create a new instance of the progress tracker with default Progres State implementation attached."""
        return SimpleProgressTracker[ProgressState](ProgressState())



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import asyncio
from typing import Any

from taskiq import TaskiqScheduler
from taskiq.cli.scheduler.args import SchedulerArgs
from taskiq.cli.scheduler.run import run_scheduler

from unikit.contrib.django.taskiq.management.base import BaseTaskiqCommand


class Command(BaseTaskiqCommand):
    """Start Taskiq worker process."""

    help = "Starts Taskiq scheduler process"

    def handle(self, *args: Any, **options: Any) -> None:
        """Start Taskiq worker process."""
        broker_name = self._get_broker_name(args)
        args = tuple(["<PLACEHOLDER>"] + list(args))

        scheduler_args = SchedulerArgs.from_cli(args)
        scheduler_args.tasks_pattern = self.taskiq_app.task_discovery_pattern

        from taskiq.schedule_sources import LabelScheduleSource

        broker = self.taskiq_app.brokers.get(broker_name)
        if broker is None:
            raise ValueError(f"Broker {broker_name} is not found.")

        scheduler = TaskiqScheduler(
            broker=broker,
            sources=[LabelScheduleSource(broker)],
        )
        scheduler_args.scheduler = scheduler

        if not scheduler_args.modules:
            scheduler_args.fs_discover = True

        asyncio.run(run_scheduler(scheduler_args))



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from typing import Any, Generic, cast, overload

from taskiq import Context
from taskiq.depends.progress_tracker import TaskProgress, TaskState
from taskiq.kicker import _T
from taskiq_dependencies import Depends

from unikit.progress import ProgressState, ProgressTracker, TProgressState
from unikit.utils import dict_utils
from unikit.utils.default import OnErrorDef, raise_or_default
from unikit.worker import RESULT_KEY_PROGRESS_STATE


class TaskiqProgressTracker(ProgressTracker[TProgressState], Generic[TProgressState]):
    """Implementation of the ProgressTracker which reports progress to Taskiq."""

    def __init__(
        self,
        progress_reporter: "TaskProgressReporter",
        progress_state: TProgressState,
        report_every_x_updates: int = 5,
    ):
        super().__init__(progress_state, report_every_x_updates)
        self._progress_reporter = progress_reporter
        self._state_cls = self.state.__class__

    async def init(self) -> None:
        """Initialize asynchronous components."""
        await self.afetch_state()

    async def afetch_state(self) -> TProgressState:
        """Fetch state from the backend."""
        self._state = await self._progress_reporter.get_object(
            self._state_cls, key=RESULT_KEY_PROGRESS_STATE, on_missing=self._state
        )
        return self.state

    def fetch_state(self) -> TProgressState:
        """NOT IMPLEMENTED FOR TASKIQ."""
        raise NotImplementedError("Taskiq doesn't support sync operations")

    def _do_report_update(self) -> None:
        """NOT IMPLEMENTED FOR TASKIQ."""
        raise NotImplementedError("Taskiq doesn't support sync operations")

    async def _ado_report_update(self) -> None:
        await self._progress_reporter.set_object(self.state, key=RESULT_KEY_PROGRESS_STATE)
        await self.afetch_state()


class TaskProgressReporter:
    """Task's dependency to set progress."""

    def __init__(
        self,
        context: Context = Depends(),
    ) -> None:
        self.context = context

    async def set_object(self, *objects: Any, key: str | None = None) -> None:
        """Set object to progress."""
        progress = await self.get_progress()
        if progress is None:
            progress = TaskProgress(state=TaskState.STARTED, meta={})
        if progress.meta is None:
            progress.meta = {}
        dict_utils.set_objects(progress.meta, *objects, key=key)
        await self.context.broker.result_backend.set_progress(
            self.context.message.task_id,
            progress,
        )

    async def get_object(self, target_cls: type[_T], key: str | None = None, on_missing: OnErrorDef[Any] = None) -> Any:
        """Get object from progress."""
        progress = await self.get_progress()
        if progress is None:
            return raise_or_default(on_missing, "Progress is not available")
        if progress.meta is None:
            progress.meta = {}
        return dict_utils.get_object(progress.meta, target_cls, key=key, on_missing=on_missing)

    async def set_progress(
        self,
        state: TaskState | str,
        meta: dict[str, Any] | None = None,
    ) -> None:
        """Set progress."""
        if meta is None:
            progress = await self.get_progress()
            meta = progress.meta if progress else None

        progress = TaskProgress(state=state, meta=meta)

        await self.context.broker.result_backend.set_progress(
            self.context.message.task_id,
            progress,
        )

    async def get_progress(self) -> TaskProgress[dict[str, Any]] | None:
        """Get progress."""
        return await self.context.broker.result_backend.get_progress(
            self.context.message.task_id,
        )

    @overload
    async def create_tracker(self, *, report_every_x_updates: int = 5) -> TaskiqProgressTracker[ProgressState]:
        pass

    @overload
    async def create_tracker(
        self, progress_state: TProgressState, report_every_x_updates: int = 5
    ) -> TaskiqProgressTracker[TProgressState]:
        pass

    async def create_tracker(
        self, progress_state: TProgressState | None = None, report_every_x_updates: int = 5
    ) -> TaskiqProgressTracker[TProgressState]:
        """Create a new progress tracker."""
        actual_progress_state: TProgressState | ProgressState
        if progress_state is None:
            actual_progress_state = ProgressState()
        else:
            actual_progress_state = progress_state
        tracker = TaskiqProgressTracker(self, actual_progress_state, report_every_x_updates)
        await tracker.init()
        return cast(TaskiqProgressTracker[TProgressState], tracker)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
from contextvars import ContextVar
import logging
from typing import Any, Self, TypeVar

from injector import Injector, Provider

from unikit.abstract import Abstract, AbstractMeta
from unikit.di import DiProxy, root_container


class SecurityContextDto(dict[str, Any]):
    """Dto representing serializable SecurityContext."""

    @property
    def principal_id(self) -> str | None:
        """Get principal ID."""
        return self.get(BaseSecurityContext.KEY_PRINCIPAL_ID)

    @classmethod
    def from_dict(cls, data: dict[str, Any] | None) -> Self:
        """Create SecurityContextDto from dictionary."""
        return cls(data or {})


class BaseSecurityContext(Abstract, metaclass=AbstractMeta):
    """Base class for Security Context."""

    KEY_PRINCIPAL_ID = "principal_id"

    @abc.abstractmethod
    def get_principal_id(self) -> str | None:
        """Get principal ID."""
        pass

    @classmethod
    @abc.abstractmethod
    def from_dto(cls, dto: SecurityContextDto) -> Self:
        """Create Security Context from serializable dto."""
        pass

    @classmethod
    @abc.abstractmethod
    async def afrom_dto(cls, dto: SecurityContextDto) -> Self:
        """Create Security Context from serializable dto."""
        pass

    def to_dto(self) -> SecurityContextDto:
        """Convert Security Context to serializable dictionary."""
        return SecurityContextDto({self.KEY_PRINCIPAL_ID: self.get_principal_id()})

    def __str__(self) -> str:
        return f"SecurityContext(principal_id={self.get_principal_id()})"


TBaseSecurityContext = TypeVar("TBaseSecurityContext", bound=BaseSecurityContext)


class SecurityContextProxy(DiProxy[BaseSecurityContext]):
    """DiProxy for Security Context."""

    pass


class ContextVarSecurityContextHolder:
    """
    A holder which utilizes ContextVar to store Security Context.

    It can be used to store security context in coroutine bound container. This can be useful to make Security Context
    available across your workers for example.

    Typical usage with DI:
        class _Module(DiModule):

            def configure(self, binder: injector.Binder):
                super().configure(binder)
                self.register_security_context(MySecurityContextInterface, ContextVarSecurityContextProvider)
    """

    def __init__(self) -> None:
        self.__security_context = ContextVar[BaseSecurityContext | None]("_unikit__security_context", default=None)

    def set_security_context(self, security_context: BaseSecurityContext | None) -> None:
        """Set Security Context."""
        self.__security_context.set(security_context)

    def get_security_context(self) -> BaseSecurityContext | None:
        """Get Security Context."""
        return self.__security_context.get()


default_security_context_holder = ContextVarSecurityContextHolder()


class ContextVarSecurityContextProvider(Provider[BaseSecurityContext]):
    """
    Provider for Security Context based on ContextVarSecurityContextHolder.

    Target context holder must be passed in constructor. If not passed, default_security_context_holder will be used.
    """

    def __init__(self, context_holder: ContextVarSecurityContextHolder | None = None):
        super().__init__()
        self._context_holder = context_holder or default_security_context_holder

    def get(self, injector: Injector) -> BaseSecurityContext:
        """Get Web Security Context."""
        security_context = self._context_holder.get_security_context()
        if security_context is None:
            raise ValueError("Security Context is not available.")
        return security_context


class SecurityContextLoggingFilter(logging.Filter):
    """Logging filter to attach Security Context entries to every log record."""

    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter log record."""
        try:
            security_context = root_container.get(BaseSecurityContext)  # type: ignore[type-abstract]
            if security_context:
                dto = security_context.to_dto()
                for k, v in dto.items():
                    setattr(record, self.prefix + k, v)
        except Exception:
            pass  # This filter should never raise an exception
        return True



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import datetime
from typing import Any

from asgiref.sync import async_to_sync
from taskiq import AsyncBroker, AsyncTaskiqTask, TaskiqResult
from taskiq.depends.progress_tracker import TaskProgress
from taskiq.kicker import AsyncKicker

from unikit.contrib.taskiq.dto import (
    TASKIQ_LABEL_DATE_POSTED,
    TASKIQ_LABEL_SECURITY_CONTEXT,
    TASKIQ_LABEL_TASKNAME,
    TaskiqPostedTask,
)
from unikit.security_context import SecurityContextDto
from unikit.worker import JobStatus, PostedTask, TaskResult, WorkerService


class TaskiqWorkerService(WorkerService):
    """Implementation of WorkerService for Taskiq."""

    def __init__(self, broker: AsyncBroker):
        super().__init__()
        self.broker = broker

    async def aget_task_result(self, job_uuid: str) -> TaskResult:
        """Get task result by UUID."""
        taskiq_result: TaskiqResult | None = None
        try:
            taskiq_result = await self.broker.result_backend.get_result(job_uuid, with_logs=True)
        except Exception:
            progress = await self.broker.result_backend.get_progress(job_uuid)
            if progress:
                return self.__taskiq_progress_to_task_result(job_uuid, progress)
        if taskiq_result is None:
            return TaskResult(uuid=job_uuid, status=JobStatus.PENDING)
        return self.__taskiq_result_to_task_result(job_uuid, taskiq_result)

    async def await_for_task(self, job_uuid: str, timeout: datetime.timedelta | None = None) -> TaskResult:
        """Wait for task by UUID."""
        taskiq_result = await AsyncTaskiqTask(
            task_id=job_uuid,
            result_backend=self.broker.result_backend,
        ).wait_result(timeout=timeout.seconds if timeout else -1.0)
        return self.__taskiq_result_to_task_result(job_uuid, taskiq_result)

    async def apost_task(self, name: str, *args: Any, **kwargs: Any) -> PostedTask:
        """Post task by name."""
        try:
            task = self.broker.find_task(name)
            if task is None:
                raise ValueError(f"Task {name} is not a known task.")
            kicker = task.kicker()
        except ValueError:
            kicker = AsyncKicker(
                task_name=name,
                broker=self.broker,
                labels={},
            )

        kicked_task = await kicker.with_broker(self.broker).kiq(*args, **kwargs)
        return TaskiqPostedTask(
            uuid=kicked_task.task_id, timestamp=datetime.datetime.now(), task=kicked_task, task_name=name
        )

    def get_task_result(self, job_uuid: str) -> TaskResult:
        """Get task result by UUID."""
        return async_to_sync(self.aget_task_result)(job_uuid)

    def wait_for_task(self, job_uuid: str, timeout: datetime.timedelta | None = None) -> TaskResult:
        """Wait for task by UUID."""
        return async_to_sync(self.await_for_task)(job_uuid, timeout)

    def post_task(self, name: str, *args: Any, **kwargs: Any) -> PostedTask:
        """Post task by name."""
        return async_to_sync(self.apost_task)(name, *args, **kwargs)

    def supports_task(self, task_name: str) -> bool:
        """Check if the worker service supports the given task."""
        return self.broker.find_task(task_name) is not None

    def __taskiq_progress_to_task_result(self, uuid: str, progress: TaskProgress[dict[str, Any]]) -> TaskResult:
        labels = progress.meta.get("labels", {}) if progress.meta else {}
        time_posted_str = labels.get(TASKIQ_LABEL_DATE_POSTED)
        time_posted = datetime.datetime.fromisoformat(time_posted_str) if time_posted_str else None
        return TaskResult(
            uuid=uuid,
            status=JobStatus.RUNNING,
            result=progress.meta,
            timestamp=time_posted,
            task_name=labels.get(TASKIQ_LABEL_TASKNAME),
            security_context=SecurityContextDto.from_dict(labels.get(TASKIQ_LABEL_SECURITY_CONTEXT)),
        )

    def __taskiq_result_to_task_result(self, uuid: str, result: TaskiqResult) -> TaskResult:
        time_posted_str = result.labels.get(TASKIQ_LABEL_DATE_POSTED)
        time_posted = datetime.datetime.fromisoformat(time_posted_str) if time_posted_str else None
        return TaskResult(
            uuid=uuid,
            status=JobStatus.FAILED if result.is_err else JobStatus.SUCCESS,
            result=result.return_value,
            duration=datetime.timedelta(seconds=result.execution_time),
            timestamp=time_posted,
            log=result.log,
            error_message=str(result.error) if result.error else None,
            task_name=result.labels.get(TASKIQ_LABEL_TASKNAME),
            security_context=SecurityContextDto.from_dict(result.labels.get(TASKIQ_LABEL_SECURITY_CONTEXT)),
        )



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
from contextvars import ContextVar
import dataclasses
import datetime
from enum import StrEnum
import logging
from typing import Any

from injector import Injector, Provider

from unikit.abstract import Abstract, AbstractMeta
from unikit.di import root_container
from unikit.progress import ProgressState
from unikit.registry import Registry
from unikit.security_context import SecurityContextDto
from unikit.utils import dict_utils
from unikit.utils.default import OnErrorDef, raise_or_default
from unikit.utils.type_utils import T

RESULT_KEY_PROGRESS_STATE = "progress_state"


class JobStatus(StrEnum):
    """Enumeration of possible job statuses."""

    PENDING = "PENDING"
    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"
    FAILED = "FAILED"


@dataclasses.dataclass(kw_only=True, frozen=True)
class PostedTask:
    """DTO for posted task."""

    uuid: str
    timestamp: datetime.datetime
    task_name: str | None = None


@dataclasses.dataclass(kw_only=True)
class TaskResult:
    """DTO for task result."""

    uuid: str
    status: JobStatus
    result: Any = None
    timestamp: datetime.datetime | None = None
    error_message: str | None = None
    log: str | None = None
    duration: datetime.timedelta | None = None
    task_name: str | None = None
    security_context: SecurityContextDto = dataclasses.field(default_factory=SecurityContextDto)

    def get_result_obj(self, target_cls: type[T], key: str | None = None, on_missing: OnErrorDef[Any] = None) -> Any:
        """Get the result object by the given target class."""
        if self.result is None:
            return raise_or_default(on_missing, "Result is not available")
        return dict_utils.get_object(self.result, target_cls, key=key, on_missing=on_missing)

    def get_progress_state(self) -> ProgressState:
        """Get the progress state from the result."""
        default = ProgressState()
        return self.get_result_obj(ProgressState, key=RESULT_KEY_PROGRESS_STATE, on_missing=default)

    @property
    def is_progress_state_available(self) -> bool:
        """Check if the progress state is available."""
        return not self.get_progress_state().is_empty


class WorkerService(Abstract, metaclass=AbstractMeta):
    """Worker service interface."""

    def prepare_for_serialization(self, obj: Any) -> Any:
        """Prepare the object for serialization."""
        if hasattr("to_dict", obj):
            return obj.to_dict()
        elif hasattr("serialize", obj):
            return obj.serialize()
        else:
            return obj

    @abc.abstractmethod
    def get_task_result(self, job_uuid: str) -> TaskResult:
        """Get the task result by the given job UUID."""
        pass

    @abc.abstractmethod
    async def aget_task_result(self, job_uuid: str) -> TaskResult:
        """Get the task result by the given job UUID asynchronously."""
        pass

    @abc.abstractmethod
    def wait_for_task(self, job_uuid: str, timeout: datetime.timedelta | None = None) -> TaskResult:
        """Wait for the task by the given job UUID."""
        pass

    @abc.abstractmethod
    async def await_for_task(self, job_uuid: str, timeout: datetime.timedelta | None = None) -> TaskResult:
        """Wait for the task by the given job UUID asynchronously."""
        pass

    @abc.abstractmethod
    def post_task(self, name: str, *args: Any, **kwargs: Any) -> PostedTask:
        """Post the task by the given name."""
        pass

    @abc.abstractmethod
    async def apost_task(self, name: str, *args: Any, **kwargs: Any) -> PostedTask:
        """Post the task by the given name asynchronously."""
        pass

    @abc.abstractmethod
    def supports_task(self, task_name: str) -> bool:
        """
        Check if the worker service supports the given task.

        :param task_name: name of the task
        :return: True if the worker service supports the task, False otherwise
        """
        pass


class WorkerServiceRegistry(Registry[str, WorkerService]):
    """A registry for WorkerService objects."""

    def get_for_task(self, task_name: str) -> WorkerService:
        """
        Get the worker service for the given task name.

        :param task_name: name of the task
        :return: worker service for the task
        """
        for ws in self.get_all():
            ws.supports_task(task_name)
            return ws
        raise KeyError(f"Worker service for task `{task_name}` not found in registry.")

    def get_for_task_or_default(self, task_name: str) -> WorkerService:
        """Get the worker service for the given task name or default."""
        try:
            return self.get_for_task(task_name)
        except KeyError:
            return self.get_default()

    def get_default(self) -> WorkerService:
        """
        Get the default worker service.

        :return: default worker service
        """
        return next(self.get_all())

    def get_task_result(self, job_uuid: str) -> TaskResult:
        """Get the task result by the given job UUID."""
        return self.get_default().get_task_result(job_uuid)

    async def aget_task_result(self, job_uuid: str) -> TaskResult:
        """Get the task result by the given job UUID asynchronously."""
        return await self.get_default().aget_task_result(job_uuid)

    def wait_for_task(self, job_uuid: str, timeout: datetime.timedelta | None = None) -> TaskResult:
        """Wait for the task by the given job UUID."""
        return self.get_default().wait_for_task(job_uuid, timeout)

    async def await_for_task(self, job_uuid: str, timeout: datetime.timedelta | None = None) -> TaskResult:
        """Wait for the task by the given job UUID asynchronously."""
        return await self.get_default().await_for_task(job_uuid, timeout)

    def post_task(self, name: str, *args: Any, fallback_to_default: bool = True, **kwargs: Any) -> PostedTask:
        """Post the task by the given name."""
        service = self.get_for_task_or_default(name) if fallback_to_default else self.get_for_task(name)
        return service.post_task(name, *args, **kwargs)

    async def apost_task(self, name: str, *args: Any, fallback_to_default: bool = True, **kwargs: Any) -> PostedTask:
        """Post the task by the given name asynchronously."""
        service = self.get_for_task_or_default(name) if fallback_to_default else self.get_for_task(name)
        return await service.apost_task(name, *args, **kwargs)


class ContextVarWorkerTaskHolder:
    """A holder which utilizes ContextVar to store current worker task."""

    def __init__(self) -> None:
        self.__current_task = ContextVar[PostedTask | None]("_unikit__current_task", default=None)

    def set_current_task(self, task: PostedTask | None) -> None:
        """Set the current worker task."""
        self.__current_task.set(task)

    def get_current_task(self) -> PostedTask | None:
        """Get the current worker task."""
        return self.__current_task.get()


class ContextVarWorkerTaskProvider(Provider[PostedTask]):
    """
    Provider for Security Context based on ContextVarSecurityContextHolder.

    Target context holder must be passed in constructor. If not passed, default_security_context_holder will be used.
    """

    def __init__(self, context_holder: ContextVarWorkerTaskHolder):
        super().__init__()
        self._context_holder = context_holder

    def get(self, injector: Injector) -> PostedTask:
        """Get Web Security Context."""
        current_task = self._context_holder.get_current_task()
        if current_task is None:
            raise ValueError("Current task is not available.")
        return current_task


class TaskInfoLoggingFilter(logging.Filter):
    """Logging filter to attach information about currently running task to every log record."""

    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter log record."""
        try:
            task: PostedTask = root_container.get(PostedTask)
            if task:
                setattr(record, self.prefix + "task_uuid", task.uuid)
                if task.task_name:
                    setattr(record, self.prefix + "task_name", task.task_name)
                if task.timestamp:
                    setattr(record, self.prefix + "task_placed_at", task.timestamp)
        except Exception:
            pass  # This filter should never raise an exception
        return True



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
__all__ = (
    "parse_bool",
    "parse_int",
    "parse_float",
    "parse_path",
    "no_trailing_slash",
    "parse_list_of_strings",
)

from pathlib import Path
from typing import overload

from .default import OnErrorDef, raise_or_default


def parse_bool(val: str | bool | int, on_error: OnErrorDef[bool] = False) -> bool:
    """
    Convert input value to the boolean if possible.

    Supported input types are `str`, `int`, `bool`.
    Non-supported input type is considered as error.
    """
    if isinstance(val, bool):
        return val
    if isinstance(val, str):
        return val.lower() in ("y", "yes", "true", "1")
    if isinstance(val, int):
        return val != 0
    return raise_or_default(on_error, f"Unable to parse value `{val}`")


@overload
def parse_int(val: str | int | None, on_error: type[Exception]) -> int: ...


@overload
def parse_int(val: str | int | None, on_error: int | None = None) -> int | None: ...


def parse_int(val: str | int | None, on_error: OnErrorDef[int | None] = None) -> int | None:
    """
    Convert given input value into integer if possible. Supported inputs are: str representing integer, integer.

    Non-supported input type is considered as error.
    :param val: input value
    :param on_error: default value or exception to be raised
    :return: int value
    """
    if val is not None:
        try:
            return int(val)
        except ValueError:
            pass
    return raise_or_default(on_error, f"Unable parse value `{val}`")


@overload
def parse_float(val: str | float | int | None, on_error: type[Exception]) -> float: ...


@overload
def parse_float(val: str | float | int | None, on_error: int | None = None) -> float | None: ...


def parse_float(val: str | float | int | None, on_error: OnErrorDef[int | None] = None) -> float | None:
    """
    Convert given input value into float if possible.

    Supported inputs are: str representing float, integer, float.
    Non-supported input type is considered as error.
    :param val: input value
    :param on_error: default value or exception to be raised
    :return: float value
    """
    if val is not None:
        try:
            return float(val)
        except ValueError:
            pass
    return raise_or_default(on_error, f"Unable parse value `{val}`")


def parse_path(val: str | Path | None, on_error: OnErrorDef[Path | None] = None) -> Path | None:
    """
    Convert given input value into Path if possible. Supported inputs are: str representing Path, Path.

    Non-supported input type is considered as error.
    :param val: input value
    :param on_error: default value or exception to be raised
    :return: Path value
    """
    if val is not None:
        try:
            return Path(val)
        except ValueError:
            pass
    return raise_or_default(on_error, f"Unable parse path from `{val}`")


def no_trailing_slash(val: str, on_error: OnErrorDef[str] = ValueError) -> str:
    """
    Remove all trailing slashes from the given string. Might be useful to normalize URLs.

    Non-string input is considered as error.
    """
    if isinstance(val, str):
        while val.endswith("/"):
            val = val[:-1]
        return val
    return raise_or_default(on_error, f"no_trailing_slash expects string but {type(val)} given")


@overload
def parse_list_of_strings(val: None, separator: str = ...) -> None: ...


@overload
def parse_list_of_strings(val: list[str] | str, separator: str = ...) -> list[str]: ...


def parse_list_of_strings(val: list[str] | str | None, separator: str = ",") -> list[str] | None:
    """
    Split comma-separated string and return it as list of values.

    :param val: input value
    :param separator: separator
    :return: list of values
    """
    if isinstance(val, str):
        return [v.strip() for v in val.split(separator)]
    return val



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
import datetime
import inspect
import json
import random
from typing import Any

from unikit.lock.service import LockService
from unikit.utils.type_utils import none_raises

from .dto import CHARACTERS, DEFAULT_LOCK_TIMEOUT, RedisLock


class BaseRedisLockService(LockService, metaclass=abc.ABCMeta):
    """Base class for Redis lock services."""

    @classmethod
    def _parse_lock(cls, lock_obj_str: str | None, keep_secret: bool = False) -> RedisLock | None:
        if lock_obj_str is None:
            return None
        try:
            lock_obj: dict[str, Any] = json.loads(lock_obj_str)
            if "s" in lock_obj and not keep_secret:
                del lock_obj["s"]
        except Exception as e:
            raise ValueError(f"Invalid redis lock object: {e}") from e

        return RedisLock.from_dict(lock_obj)

    @classmethod
    def _random_secret(cls) -> str:
        return "".join(random.choice(CHARACTERS) for _ in range(22))

    @classmethod
    def _build_new_lock(
        cls,
        op_name: str | None = None,
        target: str | None = None,
        op_params: dict[str, Any] | None = None,
        timeout: datetime.timedelta | None = None,
    ) -> RedisLock:
        timeout = timeout or DEFAULT_LOCK_TIMEOUT
        assert op_name or target, "Either op_name or target must be specified"
        target_id = cls._create_target_id(target)
        full_op_name = cls._create_op_name(op_name, op_params)
        lock_id = cls._create_pk_for_lock(full_op_name, target_id)
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        lock = RedisLock(
            lock_id=lock_id,
            op_name=op_name,
            target_id=target_id,
            ts_acquired=now,
            ts_expires=now + timeout,
            lock_owner=cls._get_lock_owner(none_raises(inspect.currentframe()).f_back, cls._get_own_files_list()),
            secret=cls._random_secret(),
        )
        return lock

    @classmethod
    def _get_own_files_list(cls) -> tuple[str, ...]:
        return super()._get_own_files_list() + (__file__,)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from unikit.contrib.django.di._injection import _DjangoInjectionApp


class DjangoDiAppConfig(_DjangoInjectionApp):
    """Django DI app enables DI django lifecycle (views, template processors, etc)."""

    name = "unikit.contrib.django.di"
    label = "django_di"
    default = True



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from collections.abc import Iterable
from typing import Any

from django.contrib.admin import AdminSite, ModelAdmin
from django.contrib.admin.exceptions import AlreadyRegistered
from django.core.exceptions import ImproperlyConfigured
from django.db.models.base import ModelBase

from unikit.di import root_container


class DiSupportedAdminSite(AdminSite):  # type: ignore[misc]
    """Custom Django admin site that supports dependency injection."""

    def __init__(self, name: str = "admin") -> None:
        super().__init__(name)
        self._di_container = root_container

    def register(self, model_or_iterable: ModelBase | Iterable, admin_class: ModelAdmin = None, **options: Any) -> None:
        """Register a model with the given admin class."""
        # The reason for this override is to instantiate model admin with call with injection
        admin_class = admin_class or ModelAdmin
        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]
        for model in model_or_iterable:
            if model._meta.abstract:
                raise ImproperlyConfigured(
                    "The model %s is abstract, so it cannot be registered with admin." % model.__name__
                )

            if self.is_registered(model):
                registered_admin = str(self.get_model_admin(model))
                msg = "The model %s is already registered " % model.__name__
                if registered_admin.endswith(".ModelAdmin"):
                    # Most likely registered without a ModelAdmin subclass.
                    msg += "in app %r." % registered_admin.removesuffix(".ModelAdmin")
                else:
                    msg += "with %r." % registered_admin
                raise AlreadyRegistered(msg)

            if not model._meta.swapped:
                if options:
                    options["__module__"] = __name__
                    admin_class = type("%sAdmin" % model.__name__, (admin_class,), options)

                # Instantiate the admin class to save in the registry
                self._registry[model] = self._di_container.create_object(
                    admin_class, additional_kwargs=dict(model=model, admin_site=self)
                )



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
import inspect
from typing import Any, Awaitable, Callable, Coroutine

from unikit.registry import T
from unikit.utils.type_utils import R

_HAS_ASGIREF = True
try:
    from asgiref.sync import async_to_sync as asgiref_async_to_sync
    from asgiref.sync import sync_to_async as asgiref_sync_to_async
except ImportError:
    _HAS_ASGIREF = False


async def maybe_awaitable(
    possible_coroutine: T | Coroutine[Any, Any, T] | Awaitable[T],
) -> T:
    """
    Awaits coroutine if needed.

    This function allows run function
    that may return coroutine.

    It not awaitable value passed, it
    returned immediately.

    :param possible_coroutine: some value.
    :return: value.
    """
    if inspect.isawaitable(possible_coroutine):
        return await possible_coroutine
    return possible_coroutine


def await_if_awaitable(possible_coroutine: T | Coroutine[Any, Any, T] | Awaitable[T]) -> T:
    """
    Awaits coroutine if needed.

    This function allows run function
    that may return coroutine.

    It not awaitable value passed, it
    returned immediately.

    :param possible_coroutine: some value.
    :return: value.
    """
    if inspect.isawaitable(possible_coroutine):
        if is_async_context():
            return asyncio.run(possible_coroutine)  # type: ignore
        else:

            async def _wrap() -> Any:
                return await possible_coroutine

            return async_to_sync(_wrap)()
    return possible_coroutine


def is_async_context() -> bool:
    """
    Check if current context is async.

    :return: True if async, False otherwise.
    """
    try:
        asyncio.get_running_loop()
        return True
    except RuntimeError:
        return False


def async_to_sync(async_func: Callable[..., Awaitable[R]], force_new_loop: bool = False) -> Callable[..., R]:
    """Convert an asynchronous function to a synchronous one."""

    if _HAS_ASGIREF:
        return asgiref_async_to_sync(async_func, force_new_loop=force_new_loop)

    @wraps(async_func)
    def sync_func(*args: T, **kwargs: T) -> R:
        loop = asyncio.new_event_loop() if force_new_loop else asyncio.get_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return loop.run_until_complete(async_func(*args, **kwargs))
        finally:
            loop.close()

    return sync_func


def sync_to_async(sync_func: Callable[..., R]) -> Callable[..., Awaitable[R]]:
    """Convert a synchronous function to an asynchronous one."""
    if _HAS_ASGIREF:
        return asgiref_sync_to_async(sync_func, thread_sensitive=True)

    @wraps(sync_func)
    async def async_func(*args: T, **kwargs: T) -> R:
        loop = asyncio.get_event_loop()
        # Run in executor to avoid blocking the event loop
        with ThreadPoolExecutor() as executor:
            return await loop.run_in_executor(executor, sync_func, *args, **kwargs)

    return async_func


def run_async(coroutine: Coroutine[Any, Any, R], force_new_loop: bool = False) -> R:
    """
    Run coroutine in async context.

    :param coroutine: coroutine.
    :param force_new_loop: force new loop.
    :return: result of coroutine.
    """
    if is_async_context():
        return asyncio.run(coroutine)

    loop = asyncio.new_event_loop() if force_new_loop else asyncio.get_event_loop()
    if force_new_loop:
        asyncio.set_event_loop(loop)
    return loop.run_until_complete(coroutine)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import os
import unittest

from unikit.contrib.redis.lock.async_ import AsyncRedisLockService
from unikit.contrib.redis.lock.sync import RedisLockService
from unikit.lock.service import LockService
from unikit.lock.test_lock import BaseLockServiceTest


@unittest.skipIf(os.environ.get("REDIS_CACHE_URL") is None, "Redis is not available, set REDIS_CACHE_URL env var")
class RedisLockServiceTest(BaseLockServiceTest):
    __test__ = True

    def _create_service(self) -> LockService:
        return RedisLockService(redis_url=os.environ.get("REDIS_CACHE_URL", "redis://localhost:1979/0"))


@unittest.skipIf(os.environ.get("REDIS_CACHE_URL") is None, "Redis is not available, set REDIS_CACHE_URL env var")
class AsyncRedisLockServiceTest(BaseLockServiceTest):
    __test__ = True

    def _create_service(self) -> LockService:
        return AsyncRedisLockService(redis_url=os.environ.get("REDIS_CACHE_URL", "redis://localhost:1979/0"))



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
__all__ = (
    "apply_pipeline",
    "normalize_spaces",
    "dedup_spaces",
    "strip_spaces",
    "strip_newlines",
)

import re
from typing import Callable, Sequence

_RE_WHITESPACES = re.compile(r"[\s\n]", re.IGNORECASE)
_RE_MULTIPLE_WHITESPACES = re.compile(r"\s+", re.IGNORECASE)


def normalize_spaces(input_: str) -> str:
    """Replace all whitespace characters with a single space."""
    return _RE_WHITESPACES.sub(" ", input_)


def dedup_spaces(input_: str) -> str:
    """Replace multiple spaces with a single space."""
    return _RE_MULTIPLE_WHITESPACES.sub(" ", input_)


def strip_spaces(input_: str) -> str:
    """Strip leading and trailing spaces from a string."""
    return input_.strip()


def strip_newlines(input_: str) -> str:
    """Strip newline characters from the beginning and end of the string."""
    return input_.strip("\n")


def apply_pipeline(target: str, pipeline: Sequence[Callable[[str], str]]) -> str:
    """Apply a sequence of functions to a string."""
    res = target
    for p in pipeline:
        res = p(res)
    return res
